/*
 *  EXP-T -- A Relativistic Fock-Space Multireference Coupled Cluster Program
 *  Copyright (C) 2018-2022 The EXP-T developers.
 *
 *  This file is part of EXP-T.
 *
 *  EXP-T is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  EXP-T is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with EXP-T.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  E-mail:        exp-t-program@googlegroups.com
 *  Google Groups: https://groups.google.com/d/forum/exp-t-program
 */

/*
 * All information about one-particle functions (spin-orbitals or spinors)
 * (and their symmetry).
 *
 * 2018-2021 Alexander Oleynichenko
 */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "engine.h"
#include "spinors.h"
#include "symmetry.h"
#include "options.h"
#include "utils.h"

// number of one-particle functions (e.g. spinors)
int NSPINORS;

// array of spinor attributes
spinor_attr_t *spinor_info;

// list of spinor indices -- of each symmetry
size_t n_spinor_blocks;
spinor_block_t *spinor_blocks;

spinor_block_t *spb_hp01[2][2];
spinor_block_t *spb_hp_val_t3[2][2][2];

// mapping "global spinor index -> local spinor index (in spinor block)"
int spinor_index_global2local[CC_MAX_SPINORS];

// helper functions
void spinor_labels_to_string(int idx, char *dst);


int get_num_spinors()
{
    return NSPINORS;
}


// info functions ("getters")
inline int is_active(int idx)
{
    return (spinor_info[idx].space_flags & CC_FLAG_ACTIVE) ? 1 : 0;
}


inline int set_active(int idx)
{
    spinor_info[idx].space_flags |= CC_FLAG_ACTIVE;
}


inline int is_hole(int idx)
{
    return spinor_info[idx].occ == 1;
}


void set_occupied(int idx, int occ_number)
{
    spinor_info[idx].occ = occ_number;
}


inline int is_act_hole(int idx)
{
    return is_hole(idx) && is_active(idx);
}


inline int is_inact_hole(int idx)
{
    return is_hole(idx) && !is_active(idx);
}


inline int is_particle(int idx)
{
    return !is_hole(idx);
}


inline int is_act_particle(int idx)
{
    return is_particle(idx) && is_active(idx);
}


inline int is_inact_particle(int idx)
{
    return is_particle(idx) && !is_active(idx);
}


inline int is_t3_space_spinor(int idx)
{
    return (spinor_info[idx].space_flags & CC_FLAG_T3_SPACE) ? 1 : 0;
}


void get_active_space_size(int *nacth, int *nactp)
{
    *nacth = 0;
    *nactp = 0;

    for (int i = 0; i < NSPINORS; i++) {
        if (is_act_hole(i)) {
            (*nacth)++;
        }
        else if (is_act_particle(i)) {
            (*nactp)++;
        }
    }
}


void get_active_holes_particles(int *nacth, int *nactp, moindex_t *active_holes_indices, moindex_t *active_parts_indices)
{
    *nacth = 0;
    *nactp = 0;

    for (int i = 0; i < NSPINORS; i++) {
        if (is_act_hole(i)) {
            active_holes_indices[*nacth] = i;
            (*nacth)++;
        }
        else if (is_act_particle(i)) {
            active_parts_indices[*nactp] = i;
            (*nactp)++;
        }
    }
}


void get_active_space(int sect_h, int sect_p, int *n_active, int *active_spinors)
{
    *n_active = 0;

    // TODO: refactor, separate utility function (move to spinors.c)
    int nspinors = get_num_spinors();
    for (int i = 0; i < nspinors; i++) {
        if ((is_act_hole(i) && sect_h > 0) || (is_act_particle(i) && sect_p > 0)) {
            active_spinors[*n_active] = i;
            *n_active = *n_active + 1;
        }
    }
}


double get_eps(int idx)
{
    return spinor_info[idx].eps;
}


void get_spinor_energies(size_t n, double *eps)
{
    for (size_t i = 0; i < n; i++) {
        eps[i] = spinor_info[i].eps;
    }
}


int get_num_electrons()
{
    int nelec = 0;

    for (size_t i = 0; i < NSPINORS; i++) {
        nelec += is_hole(i);
    }

    return nelec;
}


int get_num_spinors_irrep(int irrep_number)
{
    int n_spinors = 0;

    for (size_t i = 0; i < NSPINORS; i++) {
        if (spinor_info[i].repno == irrep_number) {
            n_spinors++;
        }
    }

    return n_spinors;
}


int get_num_electrons_irrep(int irrep_number)
{
    int n_occupied = 0;

    for (size_t i = 0; i < NSPINORS; i++) {
        if (spinor_info[i].repno == irrep_number && is_hole(i)) {
            n_occupied++;
        }
    }

    return n_occupied;
}


/**
 * @param occ_idx assumed to be allocated
 */
void get_spinor_indices_occupied(int *occ_idx)
{
    int j = 0;

    for (int i = 0; i < NSPINORS; i++) {
        if (is_hole(i)) {
            occ_idx[j] = i;
            j++;
        }
    }
}


int get_spinor_irrep(int idx)
{
    return spinor_info[idx].repno;
}


void get_spinor_info(int idx, int *repno, int *occ, int *active, double *eps)
{
    *repno = spinor_info[idx].repno;
    *occ = is_hole(idx);
    *active = is_active(idx);
    *eps = spinor_info[idx].eps;
}


void create_spinor_info(int n_spinors, int *irrep_numbers, double *spinor_energies, int *occ_numbers)
{
    NSPINORS = n_spinors;

    spinor_info = (spinor_attr_t *) cc_malloc(sizeof(spinor_attr_t) * NSPINORS);
    for (int i = 0; i < NSPINORS; i++) {
        spinor_info[i].repno = irrep_numbers[i];
        spinor_info[i].eps = spinor_energies[i];
        spinor_info[i].space_flags = 0;
        spinor_info[i].occ = occ_numbers[i];
    }
}


void create_spinor_blocks(int tilesize)
{
    int *sizes;
    int i, j, isp;
    int prt_lvl;
    int nspinors = get_num_spinors();

    prt_lvl = cc_opts->print_level;

    // count number of spinors in each irrep
    int num_irreps = get_num_irreps();
    sizes = (int *) cc_malloc(num_irreps * sizeof(int));
    memset(sizes, 0, num_irreps * sizeof(int));
    for (i = 0; i < get_num_spinors(); i++) {
        sizes[spinor_info[i].repno]++;
    }

    // count number of nonzero blocks
    n_spinor_blocks = 0;
    for (i = 0; i < num_irreps; i++) {
        if (sizes[i] > 0) {
            for (j = 0; j < sizes[i]; j += tilesize) {
                n_spinor_blocks++;
            }
        }
    }
    if (cc_opts->print_level >= CC_PRINT_DEBUG) {
        printf("nsym = %d\n", num_irreps);
        printf("sizes = ");
        for (i = 0; i < num_irreps; i++) { printf("%d ", sizes[i]); }
        printf("\n");
        printf("number of nonzero blocks = %d\n", n_spinor_blocks);
    }

    // allocate memory
    spinor_blocks = (spinor_block_t *) cc_malloc(n_spinor_blocks * sizeof(spinor_block_t));
    int k;
    j = 0;
    for (i = 0; i < num_irreps; i++) {
        if (sizes[i] > 0) {
            for (k = 0; k < sizes[i]; k += tilesize) {
                int size = (k + tilesize > sizes[i]) ? (sizes[i] - k) : tilesize;
                if (cc_opts->print_level >= CC_PRINT_DEBUG) {
                    printf("i = %d j = %d repno = %d size = %d\n", i, j, i, size);
                }
                spinor_blocks[j].size = 0;
                spinor_blocks[j].repno = i;
                spinor_blocks[j].indices = (int *) cc_malloc(size * sizeof(int));
                j++;
            }
        }
    }

    // assign indices
    for (i = 0; i < nspinors; i++) {
        for (j = 0; j < n_spinor_blocks; j++) {
            if (spinor_info[i].repno == spinor_blocks[j].repno) {
                if (spinor_blocks[j].size == tilesize) {
                    continue;
                }
                spinor_blocks[j].indices[spinor_blocks[j].size++] = i;
                spinor_info[i].blockno = j;
                break;
            }
        }
    }

    int field_width = 7;
    for (int i = 0; i < n_spinor_blocks; i++) {
        int len = strlen(get_irrep_name(spinor_info[i].repno));
        if (len > field_width) {
            field_width = len;
        }
    }

    /* beautiful table with information about spinor blocks */
    if (cc_opts->print_level >= CC_PRINT_HIGH) {
        printf(" Blocks of molecular spinors:\n");
        print_hyphens(1, 80);
        printf("   #  size%*s%s\n", field_width, "irrep", "  spinor indices");
        print_hyphens(1, 80);
        for (int i = 0; i < n_spinor_blocks; i++) {
            size_t sz = spinor_blocks[i].size;
            char *rep_name = get_irrep_name(spinor_blocks[i].repno);
            int *indices = spinor_blocks[i].indices;
            printf(" [%2d] %4d%*s  ", i, sz, field_width, rep_name);
            for (int j = 0; j < sz; j++) {
                if (j != 0) {
                    printf(",");
                }
                printf("%d", indices[j] + 1);

                int nskip = 0;
                for (int k = j + 1; k < sz; k++) {
                    if (indices[k] == indices[k - 1] + 1) {
                        nskip++;
                    } else {
                        break;
                    }
                }
                if (nskip > 0) {
                    printf("-%d", indices[j + nskip] + 1);
                    j += nskip;
                }
            }
            printf("\n");
        }
        print_hyphens(1, 80);
        printf("\n");
    }

    // construct mapping: "global spinor index -> local spinor index"
    for (isp = 0; isp < nspinors; isp++) {
        for (i = 0; i < n_spinor_blocks; i++) {
            for (j = 0; j < spinor_blocks[i].size; j++) {
                if (spinor_blocks[i].indices[j] == isp) {
                    spinor_index_global2local[isp] = j;
                    goto contin;
                }
            }
        }
        contin:
        continue;
    }

    if (prt_lvl >= CC_PRINT_DEBUG) {
        printf("spinor indices, global to local mapping:\n");
        for (i = 0; i < nspinors; i++) {
            printf("%3d->%3d ", i, spinor_index_global2local[i]);
            if ((i + 1) % 10 == 0) {
                printf("\n");
            }
        }
        printf("\n");
    }

    // cleanup
    cc_free(sizes);
}


int is_symblock_zero(int rank, int *spinor_blocks, int *qparts, int *valence, int *t3space)
{
    for (int i = 0; i < rank; i++) {
        int hp = (qparts[i] == 'h') ? 0 : 1;
        int val = (valence[i] == 1) ? 1 : 0;
        int t3 = (t3space[i] == 1) ? 1 : 0;

        spinor_block_t *arr = NULL;
        if (!cc_opts->do_restrict_t3) {
            arr = spb_hp01[hp][val];
        }
        else {
            arr = spb_hp_val_t3[hp][val][t3];
        }
        int bsize = arr[spinor_blocks[i]].size;
        if (bsize == 0) {
            return 1;
        }
    }
    return 0;
}


int spinor_attr_cmp_by_eps(const void *i, const void *j)
{
    double eps1 = ((const spinor_attr_t *) i)->eps;
    double eps2 = ((const spinor_attr_t *) j)->eps;

    if (fabs(eps1 - eps2) < 1e-10) {
        return 0;
    }
    if (eps1 > eps2) {
        return 1;
    }
    else {
        return -1;
    }
}


void setup_occupation_numbers(cc_options_t *options, int num_spinors, spinor_attr_t *spinor_info)
{
    if (!options->occ_defined && !options->nelec_defined) {
        // occupation numbers are imported via the interface
        return;
    }

    if (options->occ_defined) {
        for (int i = 0; i < NSPINORS; i++) {
            set_occupied(i, options->occ[i]);
        }
        return;
    }

    if (options->nelec_defined) {
        int nelec[CC_MAX_NUM_IRREPS];
        memset(nelec, 0, sizeof(nelec));

        // for each irrep: find its occupation number in the 'options' structure
        for (int i = 0; i < CC_MAX_NUM_IRREPS; i++) {
            int irrep_occ = options->irrep_occ_numbers.dim[i];
            if (irrep_occ > 0) {
                char *irrep_name = options->irrep_occ_numbers.rep_names[i];
                int irrep_number = get_rep_number(irrep_name);
                nelec[irrep_number] = irrep_occ;
            }
        }

        for (int i = 0; i < num_spinors; i++) {
            int repno = spinor_info[i].repno;
            if (nelec[repno] > 0) {
                set_occupied(i, 1);
                nelec[repno]--;
            }
            else {
                set_occupied(i, 0);
            }
        }
    }
}


void setup_active_space_by_energy(double lower_energy_limit, double upper_energy_limit, int space_flag)
{
    for (int i = 0; i < NSPINORS; i++) {
        double eps = spinor_info[i].eps;
        if (eps >= lower_energy_limit && eps <= upper_energy_limit) {
            spinor_info[i].space_flags |= space_flag;
        }
    }
}


/**
 * NOTE: spinors in the list are not guaranteed to be sorted by energy (in ascending order).
 * In order to find:
 *   nacth active holes (max in energy)
 *   nactp active particles (min in energy)
 * we must perform additional sorting of spinors by their energy.
 */
void setup_active_space_by_total(int nacth, int nactp, int space_flag)
{
    spinor_attr_t *sorted_spinor_info = (spinor_attr_t *) cc_malloc(NSPINORS * sizeof(spinor_attr_t));

    memcpy(sorted_spinor_info, spinor_info, sizeof(spinor_attr_t) * NSPINORS);
    qsort(sorted_spinor_info, NSPINORS, sizeof(spinor_attr_t), spinor_attr_cmp_by_eps);

    int num_occupied = get_num_electrons();
    double lower_energy_limit = sorted_spinor_info[num_occupied - nacth].eps - 1e-7;
    double upper_energy_limit = sorted_spinor_info[num_occupied + nactp - 1].eps + 1e-7;
    setup_active_space_by_energy(lower_energy_limit, upper_energy_limit, space_flag);

    cc_free(sorted_spinor_info);
}


void setup_active_space_by_binary_vector(int *active_vector, int space_flag)
{
    for (int i = 0; i < NSPINORS; i++) {
        if (active_vector[i] == 1) {
            spinor_info[i].space_flags |= space_flag;
        }
    }
}


void setup_active_space_by_irreps(int space_flag)
{
    // here we assume that all spinors inside the irrep are sorted by energy
    // this is true for DIRAC
    for (size_t i = 0; i < CC_MAX_NUM_IRREPS; i++) {
        cc_active_spec_t *sp = cc_opts->active_specs + i;
        if (*(sp->rep_name) == '\0') {
            continue;
        }
        int irep = get_rep_number(sp->rep_name);
        if (irep == -1) {
            errquit("Error in the input file: wrong irrep name in active space specification: '%s'",
                    sp->rep_name);
        }
        // count number of spinors and occupied spinors in this irrep
        int n_spi = get_num_spinors_irrep(irep);
        if (n_spi == 0) {
            printf(" Warning: there are no spinors belonging to the irrep '%s'\n", sp->rep_name);
            continue;
        }

        int n_occ = get_num_electrons_irrep(irep);
        int n_inact_h = n_occ - sp->nacth;

        int count = 0;
        for (size_t ispinor = 0; ispinor < NSPINORS; ispinor++) {
            spinor_attr_t *info = &spinor_info[ispinor];
            if (info->repno != irep) {
                continue;
            }

            if (count < n_inact_h) {
                count++;
                continue;
            }
            if (count < n_occ) {
                spinor_info[ispinor].space_flags |= space_flag;
                count++;
                continue;
            }
            if (count < n_occ + sp->nactp) {
                spinor_info[ispinor].space_flags |= space_flag;
                count++;
                continue;
            }
        }
    }
}


void setup_active_space(cc_options_t *cc_opts)
{
    /* active space */
    if (cc_opts->actsp_defined == CC_ACT_SPACE_SPEC_ENERGY) {
        setup_active_space_by_energy(cc_opts->act_energy_min, cc_opts->act_energy_max, CC_FLAG_ACTIVE);
    }
    else if (cc_opts->actsp_defined == CC_ACT_SPACE_SPEC_TOTAL) { // by nactp/nacth (total)
        setup_active_space_by_total(cc_opts->nacth, cc_opts->nactp, CC_FLAG_ACTIVE);
    }
    else if (cc_opts->actsp_defined == CC_ACT_SPACE_SPEC_IRREPS) { // by nactp/nacth (irreps)
        setup_active_space_by_irreps(CC_FLAG_ACTIVE);
    }
    else if (cc_opts->actsp_defined == CC_ACT_SPACE_SPEC_BINARY) {
        setup_active_space_by_binary_vector(cc_opts->active_each, CC_FLAG_ACTIVE);
    }

    /* main & intermediate parts of active space */
    /*if (cc_opts->do_intham_imms) {
        setup_active_space_by_total(cc_opts->intham_imms_opts.main_holes, cc_opts->intham_imms_opts.main_particles, CC_FLAG_MAIN);
    }*/

    /* spinor space for which triple excitations will be allowed */
    if (cc_opts->cc_model > CC_MODEL_CCSD && cc_opts->do_restrict_t3) {
        setup_active_space_by_energy(cc_opts->restrict_t3_bounds[0], cc_opts->restrict_t3_bounds[1], CC_FLAG_T3_SPACE);
    }
}


spinor_block_t *get_spinors_blocks_for_space(int (*selector_fun)(int idx))
{
    spinor_block_t *selected_blocks = (spinor_block_t *) cc_malloc(sizeof(spinor_block_t) * n_spinor_blocks);

    for (size_t i = 0; i < n_spinor_blocks; i++) {
        selected_blocks[i].repno = spinor_blocks[i].repno;
        selected_blocks[i].size = 0;
        selected_blocks[i].indices = (int *) cc_malloc(spinor_blocks[i].size * sizeof(int));
        for (int j = 0; j < spinor_blocks[i].size; j++) {
            int idx = spinor_blocks[i].indices[j];
            if (selector_fun(idx)) {
                selected_blocks[i].indices[selected_blocks[i].size++] = idx;
            }
        }
    }

    return selected_blocks;
}


int is_hole_t3(int idx)
{
    return (is_hole(idx) && is_t3_space_spinor(idx)) ? 1 : 0;
}

int is_particle_t3(int idx)
{
    return (is_particle(idx) && is_t3_space_spinor(idx)) ? 1 : 0;
}

int is_act_hole_t3(int idx)
{
    return (is_act_hole(idx) && is_t3_space_spinor(idx)) ? 1 : 0;
}

int is_act_particle_t3(int idx)
{
    return (is_act_particle(idx) && is_t3_space_spinor(idx)) ? 1 : 0;
}


void setup_fast_access_spinor_lists()
{
    if (!cc_opts->do_restrict_t3) {
        spb_hp01[0][0] = get_spinors_blocks_for_space(is_hole);
        spb_hp01[0][1] = get_spinors_blocks_for_space(is_act_hole);
        spb_hp01[1][0] = get_spinors_blocks_for_space(is_particle);
        spb_hp01[1][1] = get_spinors_blocks_for_space(is_act_particle);
    }
    else {
        spb_hp_val_t3[0][0][0] = get_spinors_blocks_for_space(is_hole);
        spb_hp_val_t3[0][0][1] = get_spinors_blocks_for_space(is_hole_t3);
        spb_hp_val_t3[0][1][0] = get_spinors_blocks_for_space(is_act_hole);
        spb_hp_val_t3[0][1][1] = get_spinors_blocks_for_space(is_act_hole_t3);
        spb_hp_val_t3[1][0][0] = get_spinors_blocks_for_space(is_particle);
        spb_hp_val_t3[1][0][1] = get_spinors_blocks_for_space(is_particle_t3);
        spb_hp_val_t3[1][1][0] = get_spinors_blocks_for_space(is_act_particle);
        spb_hp_val_t3[1][1][1] = get_spinors_blocks_for_space(is_act_particle_t3);
    }
}


void print_spinor_info_table()
{
    int occ_i[CC_MAX_NUM_IRREPS];
    int occ_a[CC_MAX_NUM_IRREPS];
    int virt_i[CC_MAX_NUM_IRREPS];
    int virt_a[CC_MAX_NUM_IRREPS];
    int num_irreps = get_num_irreps();
    char spaces_label[8];

    // set counters to zero
    for (int i = 0; i < num_irreps; i++) {
        occ_i[i] = 0;
        occ_a[i] = 0;
        virt_a[i] = 0;
        virt_i[i] = 0;
    }

    printf("\n");
    printf("\t\t\t\tSpinors info\n");
    printf("\t\t\t\t------------\n");
    printf("\n");
    printf("     no    rep         occ    active        one-el energy\n");
    printf("    -------------------------------------------------------\n");
    for (int i = 0; i < NSPINORS; i++) {

        spinor_labels_to_string(i, spaces_label);

        printf("    %4d%4d \"%-6s\"%4d       %-3s   %20.12f",
               i + 1, spinor_info[i].repno,
               get_irrep_name(spinor_info[i].repno), is_hole(i),
               spaces_label,
               spinor_info[i].eps);

        if (cc_opts->spinor_labels[i] != NULL) {
            printf("    %s", cc_opts->spinor_labels[i]);
        }
        printf("\n");

        // count some statistics
        int rep = spinor_info[i].repno;
        int occ = is_hole(i);
        int act = is_active(i);
        if (occ == 1 && act == 0) {
            occ_i[rep]++;
        }
        else if (occ == 1 && act == 1) {
            occ_a[rep]++;
        }
        else if (occ == 0 && act == 0) {
            virt_i[rep]++;
        }
        else if (occ == 0 && act == 1) {
            virt_a[rep]++;
        }
    }
    printf("    -------------------------------------------------------\n\n");

    // how many reps have non-zero number of spinors
    int to_be_printed[CC_MAX_NUM_IRREPS];
    for (int irep = 0; irep < num_irreps; irep++) {
        if (occ_i[irep] == 0 && occ_a[irep] == 0 && virt_a[irep] == 0 && virt_i[irep] == 0) {
            to_be_printed[irep] = 0;
        }
        else {
            to_be_printed[irep] = 1;
        }
    }

    // print little statistics
    printf("    irreps            ");
    for (int i = 0; i < num_irreps; i++) {
        if (to_be_printed[i]) { printf("%6s", get_irrep_name(i)); }
    }
    printf("\n");
    printf("    occupied inactive ");
    for (int i = 0; i < num_irreps; i++) {
        if (to_be_printed[i]) { printf("%6d", occ_i[i]); }
    }
    printf("\n");
    printf("    occupied active   ");
    for (int i = 0; i < num_irreps; i++) {
        if (to_be_printed[i]) { printf("%6d", occ_a[i]); }
    }
    printf("\n");
    printf("    virtual active    ");
    for (int i = 0; i < num_irreps; i++) {
        if (to_be_printed[i]) { printf("%6d", virt_a[i]); }
    }
    printf("\n");
    printf("    virtual inactive  ");
    for (int i = 0; i < num_irreps; i++) {
        if (to_be_printed[i]) { printf("%6d", virt_i[i]); }
    }
    printf("\n\n");
}


void spinor_labels_to_string(int idx, char *dst)
{
    int count = 0;

    strcpy(dst, "");

    // active/inactive spinor
    if (is_active(idx)) {
        strcat(dst, "a");
        count++;
    }

    // the "t3" space: only these spinors can be included into triple excitation operators
    if (is_t3_space_spinor(idx)) {
        strcat(dst, "t");
        count++;
    }

    if (count == 0) {
        strcat(dst, "-");
    }
}


int get_spinor_block_size(int spinor_block_number)
{
    assert(spinor_block_number >= 0 && spinor_block_number < n_spinor_blocks);

    return spinor_blocks[spinor_block_number].size;
}


int get_max_spinor_block_size()
{
    int max_sz = 0;

    for (size_t i = 0; i < n_spinor_blocks; i++) {
        if (spinor_blocks[i].size > max_sz) {
            max_sz = spinor_blocks[i].size;
        }
    }

    return max_sz;
}


void spinors_cleanup()
{
    size_t i;

    cc_free(spinor_info);

    for (i = 0; i < n_spinor_blocks; i++) {
        cc_free(spinor_blocks[i].indices);
    }
    cc_free(spinor_blocks);
}

